@@ -3,17 +3,78 @@ import pandas as pd
 from sklearn.metrics import auc
 from sklearn.utils.extmath import stable_cumsum
 from sklearn.utils.validation import check_consistent_length
+from sklearn.metrics import make_scorer
 
 from ..utils import check_is_binary
 
 
+def make_uplift_scorer(metric_name, treatment, **kwargs):
+    """Make uplift scorer which can be used with the same API as ``sklearn.metrics.make_scorer``.
+
+    Args:
+        metric_name (string): Name of desirable uplift metric. Raise ValueError if invalid.
+        treatment (pandas.Series): A Series from original DataFrame which
+            contains original index and treatment group column.
+        kwargs (additional arguments): Additional parameters to be passed to metric func.
+            For example: `negative_effect`, `strategy`, `k` or somtething else.
+
+    Returns:
+        scorer (callable): An uplift scorer with passed treatment variable (and kwargs, optionally) that returns a scalar score.
+
+    Raises:
+        ValueError: if `metric_name` does not present in metrics list.
+        ValueError: if `treatment` is not a pandas Series.
+
+    Example::
+
+        from sklearn.model_selection import cross_validate
+        from sklift.metrics import make_uplift_scorer
+
+        # define X_cv, y_cv, trmnt_cv and estimator
+
+        # Use make_uplift_scorer to initialize new `sklearn.metrics.make_scorer` object
+        qini_scorer = make_uplift_scorer("qini_auc_score", trmnt_cv)
+        # or pass additional parameters if necessary
+        uplift50_scorer = make_uplift_scorer("uplift_at_k", trmnt_cv, strategy='overall', k=0.5)
+
+        # Use this object in model selection functions
+        cross_validate(estimator,
+           X=X_cv,
+           y=y_cv,
+           fit_params={'treatment': trmnt_cv}
+           scoring=qini_scorer,
+        )
+    """
+    metrics_dict = {
+        'uplift_auc_score': uplift_auc_score,
+        'qini_auc_score': qini_auc_score,
+        'uplift_at_k': uplift_at_k,
+        'weighted_average_uplift': weighted_average_uplift,
+    }
+
+    if metric_name not in metrics_dict.keys():
+        raise ValueError(
+            f"'{metric_name}' is not a valid scoring value. "
+            f"List of valid metrics: {list(metrics_dict.keys())}"
+        )
+
+    if not isinstance(treatment, pd.Series):
+        raise TypeError("Expected pandas.Series in treatment vector, got %s" % type(treatment))
+
+    def scorer(y_true, uplift, treatment_value, **kwargs):
+        t = treatment_value.loc[y_true.index]
+        return metrics_dict[metric_name](y_true, uplift, t, **kwargs)
+
+    return make_scorer(scorer, treatment_value=treatment, **kwargs)
+
+
 def uplift_curve(y_true, uplift, treatment):
     """Compute Uplift curve.
 
     For computing the area under the Uplift Curve, see :func:`.uplift_auc_score`.
 
     Args:
-        y_true (1d array-like): Correct (true) target values.
+        y_true (1d array-like): Correct (true) binary target values.
         uplift (1d array-like): Predicted uplift, as returned by a model.
         treatment (1d array-like): Treatment labels.
 
@@ -35,6 +96,8 @@ def uplift_curve(y_true, uplift, treatment):
 
     check_consistent_length(y_true, uplift, treatment)
     check_is_binary(treatment)
+    check_is_binary(y_true)
+
     y_true, uplift, treatment = np.array(y_true), np.array(uplift), np.array(treatment)
 
     desc_score_indices = np.argsort(uplift, kind="mergesort")[::-1]
@@ -75,7 +138,7 @@ def perfect_uplift_curve(y_true, treatment):
     area under the Uplift Curve, see :func:`.uplift_auc_score`.
 
     Args:
-        y_true (1d array-like): Correct (true) target values.
+        y_true (1d array-like): Correct (true) binary target values.
         treatment (1d array-like): Treatment labels.
 
     Returns:
@@ -91,6 +154,7 @@ def perfect_uplift_curve(y_true, treatment):
 
     check_consistent_length(y_true, treatment)
     check_is_binary(treatment)
+    check_is_binary(y_true)
     y_true, treatment = np.array(y_true), np.array(treatment)
 
     cr_num = np.sum((y_true == 1) & (treatment == 0))  # Control Responders
@@ -111,7 +175,7 @@ def uplift_auc_score(y_true, uplift, treatment):
     the optimum Uplift Curve.
 
     Args:
-        y_true (1d array-like): Correct (true) target values.
+        y_true (1d array-like): Correct (true) binary target values.
         uplift (1d array-like): Predicted uplift, as returned by a model.
         treatment (1d array-like): Treatment labels.
 
@@ -130,6 +194,7 @@ def uplift_auc_score(y_true, uplift, treatment):
 
     check_consistent_length(y_true, uplift, treatment)
     check_is_binary(treatment)
+    check_is_binary(y_true)
     y_true, uplift, treatment = np.array(y_true), np.array(uplift), np.array(treatment)
 
     x_actual, y_actual = uplift_curve(y_true, uplift, treatment)
@@ -149,7 +214,7 @@ def qini_curve(y_true, uplift, treatment):
     For computing the area under the Qini Curve, see :func:`.qini_auc_score`.
 
     Args:
-        y_true (1d array-like): Correct (true) target values.
+        y_true (1d array-like): Correct (true) binary target values.
         uplift (1d array-like): Predicted uplift, as returned by a model.
         treatment (1d array-like): Treatment labels.
 
@@ -174,6 +239,7 @@ def qini_curve(y_true, uplift, treatment):
 
     check_consistent_length(y_true, uplift, treatment)
     check_is_binary(treatment)
+    check_is_binary(y_true)
     y_true, uplift, treatment = np.array(y_true), np.array(uplift), np.array(treatment)
 
     desc_score_indices = np.argsort(uplift, kind="mergesort")[::-1]
@@ -214,7 +280,7 @@ def perfect_qini_curve(y_true, treatment, negative_effect=True):
     For computing the area under the Qini Curve, see :func:`.qini_auc_score`.
 
     Args:
-        y_true (1d array-like): Correct (true) target values.
+        y_true (1d array-like): Correct (true) binary target values.
         treatment (1d array-like): Treatment labels.
         negative_effect (bool): If True, optimum Qini Curve contains the negative effects
             (negative uplift because of campaign). Otherwise, optimum Qini Curve will not
@@ -232,6 +298,7 @@ def perfect_qini_curve(y_true, treatment, negative_effect=True):
 
     check_consistent_length(y_true, treatment)
     check_is_binary(treatment)
+    check_is_binary(y_true)
     n_samples = len(y_true)
 
     y_true, treatment = np.array(y_true), np.array(treatment)
@@ -261,7 +328,7 @@ def qini_auc_score(y_true, uplift, treatment, negative_effect=True):
     the optimum Qini curve.
 
     Args:
-        y_true (1d array-like): Correct (true) target values.
+        y_true (1d array-like): Correct (true) binary target values.
         uplift (1d array-like): Predicted uplift, as returned by a model.
         treatment (1d array-like): Treatment labels.
         negative_effect (bool): If True, optimum Qini Curve contains the negative effects
@@ -289,6 +356,7 @@ def qini_auc_score(y_true, uplift, treatment, negative_effect=True):
     # TODO: Add Continuous Outcomes
     check_consistent_length(y_true, uplift, treatment)
     check_is_binary(treatment)
+    check_is_binary(y_true)
     y_true, uplift, treatment = np.array(y_true), np.array(uplift), np.array(treatment)
 
     if not isinstance(negative_effect, bool):
@@ -309,7 +377,7 @@ def uplift_at_k(y_true, uplift, treatment, strategy, k=0.3):
     """Compute uplift at first k observations by uplift of the total sample.
 
     Args:
-        y_true (1d array-like): Correct (true) target values.
+        y_true (1d array-like): Correct (true) binary target values.
         uplift (1d array-like): Predicted uplift, as returned by a model.
         treatment (1d array-like): Treatment labels.
         k (float or int): If float, should be between 0.0 and 1.0 and represent the proportion of the dataset
@@ -344,6 +412,7 @@ def uplift_at_k(y_true, uplift, treatment, strategy, k=0.3):
     # TODO: checker all groups is not empty
     check_consistent_length(y_true, uplift, treatment)
     check_is_binary(treatment)
+    check_is_binary(y_true)
     y_true, uplift, treatment = np.array(y_true), np.array(uplift), np.array(treatment)
 
     strategy_methods = ['overall', 'by_group']
@@ -409,7 +478,7 @@ def response_rate_by_percentile(y_true, uplift, treatment, group, strategy='over
     """Compute response rate (target mean in the control or treatment group) at each percentile.
 
     Args:
-        y_true (1d array-like): Correct (true) target values.
+        y_true (1d array-like): Correct (true) binary target values.
         uplift (1d array-like): Predicted uplift, as returned by a model.
         treatment (1d array-like): Treatment labels.
         group (string, ['treatment', 'control']): Group type for computing response rate: treatment or control.
@@ -440,7 +509,7 @@ def response_rate_by_percentile(y_true, uplift, treatment, group, strategy='over
 
     check_consistent_length(y_true, uplift, treatment)
     check_is_binary(treatment)
-
+    check_is_binary(y_true)
     group_types = ['treatment', 'control']
     strategy_methods = ['overall', 'by_group']
     
@@ -490,7 +559,7 @@ def weighted_average_uplift(y_true, uplift, treatment, strategy='overall', bins=
     Weights are sizes of the treatment group by percentile.
 
     Args:
-        y_true (1d array-like): Correct (true) target values.
+        y_true (1d array-like): Correct (true) binary target values.
         uplift (1d array-like): Predicted uplift, as returned by a model.
         treatment (1d array-like): Treatment labels.
         strategy (string, ['overall', 'by_group']): Determines the calculating strategy. Default is 'overall'.
@@ -511,6 +580,7 @@ def weighted_average_uplift(y_true, uplift, treatment, strategy='overall', bins=
 
     check_consistent_length(y_true, uplift, treatment)
     check_is_binary(treatment)
+    check_is_binary(y_true)
 
     strategy_methods = ['overall', 'by_group']
 
@@ -540,7 +610,8 @@ def weighted_average_uplift(y_true, uplift, treatment, strategy='overall', bins=
     return weighted_avg_uplift
 
 
-def uplift_by_percentile(y_true, uplift, treatment, strategy='overall', bins=10, std=False, total=False):
+def uplift_by_percentile(y_true, uplift, treatment, strategy='overall',
+                         bins=10, std=False, total=False, string_percentiles=True):
     """Compute metrics: uplift, group size, group response rate, standard deviation at each percentile.
 
     Metrics in columns and percentiles in rows of pandas DataFrame:
@@ -552,7 +623,7 @@ def uplift_by_percentile(y_true, uplift, treatment, strategy='overall', bins=10,
         - ``std_uplift`` - (optional) uplift standard deviation.
 
     Args:
-        y_true (1d array-like): Correct (true) target values.
+        y_true (1d array-like): Correct (true) binary target values.
         uplift (1d array-like): Predicted uplift, as returned by a model.
         treatment (1d array-like): Treatment labels.
         strategy (string, ['overall', 'by_group']): Determines the calculating strategy. Default is 'overall'.
@@ -568,9 +639,10 @@ def uplift_by_percentile(y_true, uplift, treatment, strategy='overall', bins=10,
         std (bool): If True, add columns with the uplift standard deviation and the response rate standard deviation.
             Default is False.
         total (bool): If True, add the last row with the total values. Default is False.
-            The total uplift is a weighted average uplift. See :func:`.weighted_average_uplift`.
+            The total uplift computes as a total response rate treatment - a total response rate control.
             The total response rate is a response rate on the full data amount.
         bins (int): Determines the number of bins (and the relative percentile) in the data. Default is 10.
+        string_percentiles (bool): type of percentiles in the index: float or string. Default is True (string).
 
     Returns:
         pandas.DataFrame: DataFrame where metrics are by columns and percentiles are by rows.
@@ -578,6 +650,7 @@ def uplift_by_percentile(y_true, uplift, treatment, strategy='overall', bins=10,
 
     check_consistent_length(y_true, uplift, treatment)
     check_is_binary(treatment)
+    check_is_binary(y_true)
 
     strategy_methods = ['overall', 'by_group']
 
@@ -602,6 +675,10 @@ def uplift_by_percentile(y_true, uplift, treatment, strategy='overall', bins=10,
     if bins >= n_samples:
         raise ValueError(f'Number of bins = {bins} should be smaller than the length of y_true {n_samples}')
 
+    if not isinstance(string_percentiles, bool):
+        raise ValueError(f'string_percentiles flag should be bool: True or False.'
+                         f' Invalid value string_percentiles: {string_percentiles}')
+
     y_true, uplift, treatment = np.array(y_true), np.array(uplift), np.array(treatment)
 
     response_rate_trmnt, variance_trmnt, n_trmnt = response_rate_by_percentile(
@@ -613,7 +690,12 @@ def uplift_by_percentile(y_true, uplift, treatment, strategy='overall', bins=10,
     uplift_scores = response_rate_trmnt - response_rate_ctrl
     uplift_variance = variance_trmnt + variance_ctrl
 
-    percentiles = [round(p * 100 / bins, 1) for p in range(1, bins + 1)]
+    percentiles = [round(p * 100 / bins) for p in range(1, bins + 1)]
+
+    if string_percentiles:
+        percentiles = [f"0-{percentiles[0]}"] + \
+            [f"{percentiles[i]}-{percentiles[i + 1]}" for i in range(len(percentiles) - 1)]
+
 
     df = pd.DataFrame({
         'percentile': percentiles,
@@ -678,3 +760,140 @@ def treatment_balance_curve(uplift, treatment, winsize):
     balance = np.convolve(treatment, np.ones(winsize), 'valid') / winsize
     idx = np.linspace(1, 100, len(balance))
     return idx, balance
+
+
+def average_squared_deviation(y_true_train, uplift_train, treatment_train, y_true_val,
+                              uplift_val, treatment_val, strategy='overall', bins=10):
+    """Compute the average squared deviation.
+
+    The average squared deviation (ASD) is a model stability metric that shows how much the model overfits
+    the training data. Larger values of ASD mean greater overfit.
+
+    Args:
+        y_true_train (1d array-like): Correct (true) target values for training set.
+        uplift_train (1d array-like): Predicted uplift for training set, as returned by a model.
+        treatment_train (1d array-like): Treatment labels for training set.
+        y_true_val (1d array-like): Correct (true) target values for validation set.
+        uplift_val (1d array-like): Predicted uplift for validation set, as returned by a model.
+        treatment_val (1d array-like): Treatment labels for validation set.
+        strategy (string, ['overall', 'by_group']): Determines the calculating strategy. Default is 'overall'.
+
+            * ``'overall'``:
+                The first step is taking the first k observations of all test data ordered by uplift prediction
+                (overall both groups - control and treatment) and conversions in treatment and control groups
+                calculated only on them. Then the difference between these conversions is calculated.
+            * ``'by_group'``:
+                Separately calculates conversions in top k observations in each group (control and treatment)
+                sorted by uplift predictions. Then the difference between these conversions is calculated
+
+        bins (int): Determines the number of bins (and the relative percentile) in the data. Default is 10.
+
+    Returns:
+        float: average squared deviation
+
+    References:
+        René Michel, Igor Schnakenburg, Tobias von Martens. Targeting Uplift. An Introduction to Net Scores.
+    """
+    check_consistent_length(y_true_train, uplift_train, treatment_train)
+    check_is_binary(treatment_train)
+
+    check_consistent_length(y_true_val, uplift_val, treatment_val)
+    check_is_binary(treatment_val)
+
+    strategy_methods = ['overall', 'by_group']
+
+    n_samples_train = len(y_true_train)
+    n_samples_val = len(y_true_val)
+    min_n_samples = min(n_samples_train, n_samples_val)
+
+    if strategy not in strategy_methods:
+        raise ValueError(
+            f'Response rate supports only calculating methods in {strategy_methods},'
+            f' got {strategy}.')
+
+    if not isinstance(bins, int) or bins <= 0:
+        raise ValueError(
+            f'Bins should be positive integer. Invalid value bins: {bins}')
+
+    if bins >= min_n_samples:
+        raise ValueError(
+            f'Number of bins = {bins} should be smaller than the length of y_true_train {n_samples_train}'
+            f'and length of y_true_val {n_samples_val}')
+
+    uplift_by_percentile_train = uplift_by_percentile(y_true_train, uplift_train, treatment_train,
+                                                      strategy=strategy, bins=bins)
+    uplift_by_percentile_val = uplift_by_percentile(y_true_val, uplift_val, treatment_val,
+                                                    strategy=strategy, bins=bins)
+    
+    return np.mean(np.square(uplift_by_percentile_train['uplift'] - uplift_by_percentile_val['uplift']))
+
+
+def max_prof_uplift(df_sorted, treatment_name, churn_name, pos_outcome, benefit, c_incentive, c_contact, a_cost=0):
+  """Compute the maximum profit generated from an uplift model decided campaign
+
+    This can be visualised by plotting plt.plot(perc, cumulative_profit)
+
+    Args:
+      df_sorted (pandas dataframe): dataframe with descending uplift predictions for each customer (i.e. highest 1st)
+      treatment_name (string): column name of treatment columm (assuming 1 = treated)   
+      churn_name (string): column name of churn column
+      pos_outcome (int or float): 1 or 0 value in churn column indicating a positive outcome (i.e. purchase = 1, whereas churn = 0)
+      benefit (int or float): the benefit of retaining a customer (e.g., the average customer lifetime value)
+      c_incentive (int or float): the cost of the incentive if a customer accepts the offer
+      c_contact (int or float): the cost of contacting a customer regardless of conversion
+      a_cost (int or float): the fixed administration cost for the campaign
+
+    Returns:
+      1d array-like: the incremental increase in x, for plotting
+      1d array-like: the cumulative profit per customer 
+
+    References:
+        Floris Devriendt, Jeroen Berrevoets, Wouter Verbeke. Why you should stop predicting customer churn and start using uplift models.
+  """
+#   VARIABLES
+
+#       n_ct0             no. people not treated
+#       n_ct1             no. people treated
+
+#       n_y1_ct0          no. people not treated with +ve outcome
+#       n_y1_ct1          no. people treated with +ve outcome
+
+#       r_y1_ct0          mean of not treated people with +ve outcome
+#       r_y1_ct1          mean of treated people with +ve outcome
+
+#       cs                cumsum() of each variable
+
+  n_ct0 = np.where(df_sorted[treatment_name] == 0, 1, 0)
+  cs_n_ct0 = pd.Series(n_ct0.cumsum())
+
+  n_ct1 = np.where(df_sorted[treatment_name] == 1, 1, 0)
+  cs_n_ct1 = pd.Series(n_ct1.cumsum())
+
+  if pos_outcome == 0:
+    n_y1_ct0 = np.where((df_sorted[treatment_name] == 0) & (df_sorted[churn_name] == 0), 1, 0)
+    n_y1_ct1 = np.where((df_sorted[treatment_name] == 1) & (df_sorted[churn_name] == 0), 1, 0)  
+
+  elif pos_outcome == 1:
+    n_y1_ct0 = np.where((df_sorted[treatment_name] == 0) & (df_sorted[churn_name] == 1), 1, 0)
+    n_y1_ct1 = np.where((df_sorted[treatment_name] == 1) & (df_sorted[churn_name] == 1), 1, 0)
+    
+  cs_n_y1_ct0 = pd.Series(n_y1_ct0.cumsum())
+  cs_n_y1_ct1 = pd.Series(n_y1_ct1.cumsum())
+
+  cs_r_y1_ct0 = (cs_n_y1_ct0 / cs_n_ct0).fillna(0)
+  cs_r_y1_ct1 = (cs_n_y1_ct1 / cs_n_ct1).fillna(0)
+
+  cs_uplift = cs_r_y1_ct1 - cs_r_y1_ct0
+
+  # Dataframe of all calculated variables
+  df = pd.concat([cs_n_ct0,cs_n_ct1,cs_n_y1_ct0,cs_n_y1_ct1, cs_r_y1_ct0, cs_r_y1_ct1, cs_uplift], axis=1)
+  df.columns = ['cs_n_ct0', 'cs_n_ct1', 'cs_n_y1_ct0', 'cs_n_y1_ct1', 'cs_r_y1_c0', 'cs_r_y1_ct1', 'cs_uplift']
+
+  x = cs_n_ct0 + cs_n_ct1
+  max = cs_n_ct0.max() + cs_n_ct1.max()
+
+  t_profit = (x * cs_uplift * benefit) - (c_incentive * x * cs_r_y1_ct1) - (c_contact * x) - a_cost
+  perc = x / max
+  cumulative_profit = t_profit / max
+
+  return perc, cumulative_profit
