@@ -1,3 +1,3 @@
 from .utils import check_is_binary
 
-__all__ = [check_is_binary]
\ No newline at end of file
+__all__ = ['check_is_binary']
\ No newline at end of file
